// This program generates a number of testing certificates.

package main

import (
	"crypto/aes"
	"crypto/cipher"
	"crypto/rsa"
	"crypto/sha256"
	"crypto/x509"
	"crypto/x509/pkix"
	"encoding/pem"
	"math/big"
	"os"
	"os/exec"
	"time"
)

func main() {
	h := sha256.New()
	h.Write([]byte("seed for deterministic randomness"))
	rand, err := NewAESRandom(h.Sum(nil)[:16])
	if err != nil {
		panic(err)
	}

	notBefore, err := time.Parse(time.UnixDate, "Mon Jan 1 10:00:00 UTC 2013")
	if err != nil {
		panic(err)
	}

	notAfter, err := time.Parse(time.UnixDate, "Mon Dec 31 10:00:00 UTC 2023")
	if err != nil {
		panic(err)
	}

	rootKey, err := rsa.GenerateKey(rand, 2048)
	if err != nil {
		panic(err)
	}
	keyToFile("root2048.key", rootKey)

	rootTemplate := x509.Certificate{
		SerialNumber: new(big.Int).SetInt64(1),
		Subject: pkix.Name{
			Organization: []string{"Acme Co"},
			CommonName:   "Root CA",
		},
		NotBefore:             notBefore,
		NotAfter:              notAfter,
		KeyUsage:              x509.KeyUsageCertSign,
		ExtKeyUsage:           []x509.ExtKeyUsage{x509.ExtKeyUsageServerAuth},
		BasicConstraintsValid: true,
		IsCA: true,
		PermittedDNSDomains:   []string{"example.com"},
	}

	derBytes, err := x509.CreateCertificate(rand, &rootTemplate, &rootTemplate, &rootKey.PublicKey, rootKey)
	if err != nil {
		panic(err)
	}
	certToFile("root2048.crt", derBytes)

	leafKey, err := rsa.GenerateKey(rand, 2048)
	if err != nil {
		panic(err)
	}
	keyToFile("leaf2048.key", leafKey)

	leafTemplate := x509.Certificate{
		SerialNumber: new(big.Int).SetInt64(3),
		Subject: pkix.Name{
			Organization: []string{"Acme Co"},
			CommonName:   "Oversized OID element",
		},
		NotBefore:             notBefore,
		NotAfter:              notAfter,
		KeyUsage:              x509.KeyUsageDigitalSignature | x509.KeyUsageKeyEncipherment,
		ExtKeyUsage:           []x509.ExtKeyUsage{x509.ExtKeyUsageServerAuth},
		BasicConstraintsValid: true,
		IsCA:     false,
		DNSNames: []string{"example.com"},
	}

	derBytes, err = x509.CreateCertificate(rand, &leafTemplate, &rootTemplate, &leafKey.PublicKey, rootKey)
	if err != nil {
		panic(err)
	}
	certToFile("leaf2048.crt", derBytes)
}

// keyToFile writes a PEM serialization of |key| to a new file called
// |filename|.
func keyToFile(filename string, key *rsa.PrivateKey) {
	file, err := os.Create(filename)
	if err != nil {
		panic(err)
	}
	defer file.Close()
	pem.Encode(file, &pem.Block{Type: "RSA PRIVATE KEY", Bytes: x509.MarshalPKCS1PrivateKey(key)})
}

// certToFile writes a PEM serialization and OpenSSL debugging dump of
// |derBytes| to a new file called |filename|.
func certToFile(filename string, derBytes []byte) {
	cmd := exec.Command("openssl", "x509", "-text", "-inform", "DER")

	file, err := os.Create(filename)
	if err != nil {
		panic(err)
	}
	defer file.Close()
	cmd.Stdout = file
	cmd.Stderr = os.Stderr

	stdin, err := cmd.StdinPipe()
	if err != nil {
		panic(err)
	}

	if err := cmd.Start(); err != nil {
		panic(err)
	}
	if _, err := stdin.Write(derBytes); err != nil {
		panic(err)
	}
	stdin.Close()
	if err := cmd.Wait(); err != nil {
		panic(err)
	}
}

// AESRandom implements a deterministic random generator using AES in counter
// mode.
type AESRandom struct {
	cipher       cipher.Block
	buf, counter [aes.BlockSize]byte
	offset       int
}

func NewAESRandom(key []byte) (*AESRandom, error) {
	cipher, err := aes.NewCipher(key)
	if err != nil {
		return nil, err
	}
	return &AESRandom{cipher: cipher}, nil
}

func (a *AESRandom) incCounter() {
	n := 1
	for i := 0; i < aes.BlockSize; i++ {
		n += int(a.counter[i])
		a.counter[i] = byte(n)
		n >>= 8
	}
}

func (a *AESRandom) Read(buf []byte) (int, error) {
	total := len(buf)

	if a.offset != 0 {
		n := copy(buf, a.buf[a.offset:])
		a.offset = (a.offset + n) % aes.BlockSize
		buf = buf[n:]
	}

	for len(buf) >= aes.BlockSize {
		a.cipher.Encrypt(buf, a.counter[:])
		a.incCounter()
		buf = buf[aes.BlockSize:]
	}

	if len(buf) > 0 {
		a.cipher.Encrypt(a.buf[:], a.counter[:])
		a.incCounter()
		a.offset = copy(buf, a.buf[:])
	}

	return total, nil
}
